
//-----------------------------------------------------------------------------
// File: PracticalPSM.fx   
//
// From NVIDIA - Perspective Shadow Maps - PracticalPSM
//
//  version modified to be similar to "SMAP"
//
//-----------------------------------------------------------------------------

texture ShadowMap;
texture ShadowColor;
texture SpotLight;


float4x4 WorldViewProj : WorldViewProj;
float4x4 WorldIT : WorldIT;
float4x4 TexTransform;
float3   LightVec;

sampler ShadowMapSampler = sampler_state
{
    Texture = <ShadowMap>;
    MinFilter = Linear;  
    MagFilter = Linear;
    MipFilter = None;
    AddressU  = Clamp;
    AddressV  = Clamp;
};

sampler ShadowColorSampler = sampler_state
{
    Texture = <ShadowColor>;
    MinFilter = Linear;  
    MagFilter = Linear;
    MipFilter = None;
    AddressU  = Clamp;
    AddressV  = Clamp;
};




// ****************************************************************************
//   GENERATE SHADOW MAP
// ****************************************************************************

void ShadowMapVS(float4 Pos				: POSITION,
				 float3 Normal			: NORMAL, 
				 out float4 oPos		: POSITION,
				 out float4 oDepth		: TEXCOORD0)
{
    oPos = mul(Pos, WorldViewProj);
    oDepth = float4(oPos.zzz, oPos.w);
}



void BlackPS(out float4 Color : COLOR)
{
    Color = float4(0,0,0,1);
}

void OutputDepthPS(		float4 iDepth	: TEXCOORD0,
				    out float4 oColor	: COLOR )
{
    float depth = iDepth.b / iDepth.a;
    oColor = float4(depth, depth, depth , 1.0);
}



technique GenHardwareShadowMap11
{
    pass P0
    {
        VertexShader = compile vs_1_1 ShadowMapVS();
        PixelShader = compile ps_1_1 BlackPS();
    }
}

technique GenHardwareShadowMap20
{
    pass P0
    {
        VertexShader = compile vs_1_1 ShadowMapVS();
        PixelShader = compile ps_2_0 OutputDepthPS();
    }
}




// ****************************************************************************
//   USE SHADOW MAP
// ****************************************************************************

void MainVS(float4 iPos			: POSITION,
			float3 iNormal		: NORMAL, 
			out float4 oPos		: POSITION,
			out float4 oPosLight: TEXCOORD0,
			out float3 oColor0	: COLOR0)
{
    float3 worldNormal = normalize(mul(iNormal, (float3x3)WorldIT));
    float ldotn = max(dot(LightVec, worldNormal), 0.0);
    oColor0 = ldotn;

    oPos = mul(iPos, WorldViewProj);

	oPosLight = mul(iPos, TexTransform);
}


float4 MainPS_HW(float4 iPosLight : TEXCOORD0, 
				 float3 iCol	  : COLOR0) : COLOR
{
    float3 shadow = tex2D(ShadowMapSampler, iPosLight).rgb;

    float3 color = ((shadow * 0.2 + 0.8) * iCol) * 0.9 + 0.1;
    return float4(color, 1.0);
}


float4 MainPS_R32F(float4 iPosLight : TEXCOORD0, 
				   float3 iCol		: COLOR0) : COLOR
{
	float2 shadowTexCoord = iPosLight.xy / iPosLight.w;
	float  shadowTestDepth = iPosLight.z / iPosLight.w;
	
	float  shadowDepth = tex2D(ShadowMapSampler, shadowTexCoord);
	float  shadow      = (shadowTestDepth <= shadowDepth);

	float3 color = ((shadow * 0.2 + 0.8) * iCol) * 0.9 + 0.1;
	return float4(color, 1.0);
}

technique UseHardwareShadowMap
{
    pass P0
    {
        VertexShader = compile vs_1_1 MainVS();
        PixelShader = compile ps_1_1 MainPS_HW();
    }
}

technique UseR32FShadowMap
{
    pass P0
    {
        VertexShader = compile vs_1_1 MainVS();
        PixelShader = compile ps_2_0 MainPS_R32F();
    }
}



// ****************************************************************************
//   DRAW SHADOW MAP
// ****************************************************************************

void QuadVS(float4 iPos		: POSITION,
			out float4 oPos : POSITION,
			out float4 oTex0: TEXCOORD0)
{
   float2 Position = (iPos.xz / 1600.0f) + 0.5f;
   oPos = float4(Position * 0.45 + 0.5, 1, 1);
   oTex0 = float4(Position.x, 1.0 - Position.y, 0, 0);
}


float4 DrawShadowPS(float4 iTex0 : TEXCOORD0) : COLOR
{
    return float4(tex2D(ShadowColorSampler, iTex0).rgb, 1.0);
}


technique DrawHardwareShadowMap
{
    Pass P0
    {
        VertexShader = compile vs_1_1 QuadVS();
        PixelShader = compile ps_1_1 DrawShadowPS();
    }
}

//-----------------------------------------------------------------------------

