#include "DXUT.h"
#include "App.h"

//#define ENABLE_VSYNC
#define DRAW_GRID
#define DRAW_AXIS
//#define CONFINE_MOUSE

#define INIT_CLIENT

//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
HRESULT Initialize( int width, int height, const WCHAR *pWindowTitle, App *app )
{	
	gApp = app;

	LOG_INFO( "Initializing DXUT..." );

	// Set DXUT callbacks
	DXUTSetCallbackMsgProc( MsgProc );
	DXUTSetCallbackKeyboard( OnKeyboard );
	DXUTSetCallbackMouse( OnMouseProc );
	DXUTSetCallbackFrameMove( OnFrameMove );
	DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
	DXUTSetCallbackD3D9DeviceAcceptable( IsD3D9DeviceAcceptable );
	DXUTSetCallbackD3D9DeviceCreated( OnD3D9CreateDevice );
	DXUTSetCallbackD3D9DeviceReset( OnD3D9ResetDevice );
	DXUTSetCallbackD3D9DeviceLost( OnD3D9LostDevice );
	DXUTSetCallbackD3D9DeviceDestroyed( OnD3D9DestroyDevice );
	DXUTSetCallbackD3D9FrameRender( OnD3D9FrameRender );
	
	LOG_INFO( "Calling OneTimeInit()" );
	gApp->OneTimeInit();

#ifdef INIT_CLIENT
	LOG_INFO( "Initializing client" );	
	gApp->OneTimeClientInit();	
#endif
	
	LOG_INFO( "Initializing Window and D3D Device..." );
	DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
	DXUTSetCursorSettings( true, true );
	DXUTCreateWindow( pWindowTitle );
	
	// Force window to show
	ShowWindow( DXUTGetHWND(), SW_SHOW);
	UpdateWindow( DXUTGetHWND() );

	DXUTCreateDevice( true, width, height );

#ifdef CONFINE_MOUSE
	gApp->ConfineMouse();
#endif
	
	ShowWindow( DXUTGetHWND(), SW_SHOW | SW_RESTORE);

	LOG_INFO( "Done intialization. Entering DXUTMainLoop()..." );	
	DXUTMainLoop(); // Enter into the DXUT render loop
	
	return DXUTGetExitCode();
}

App::App(void)
{
	LOG_INFO( "Application constructor called." );

	mAppExiting			= false;	
	mCamSpeed			= 50.0f;
	mGridPoints			= NULL;	
	mRightButtonDown	= false;	
	mFont				= NULL;
	mLine				= NULL;
	mSprite				= NULL;	
	mShowGui			= true;
	mWorstFps			= 1000.f;
	mBestFps			= 0.f;
	mAvgFps				= 0.f;	

	srand(GetTickCount());
}

App::~App(void)
{		
	LOG_INFO("Host application destructor called.");
	// Log fps
	LOG_INFO("Fps: " << DXUTGetFPS() << " Average: " << mAvgFps << " Best: " << mBestFps << " Worst: " << mWorstFps);
	SAFE_DELETE_ARRAY( mGridPoints );
}

void App::OnFrameRender( double time, float elapsedTime )
{
	D3DXMATRIX matWorld;

	// Moved to client for this demo
	//mDevice->SetTransform( D3DTS_PROJECTION, mCamera.GetProjMatrix() );
	//mDevice->SetTransform( D3DTS_VIEW, mCamera.GetViewMatrix() );

#ifdef DRAW_GRID
	RenderGrid();
#endif

}


void App::OnFrameMove( double time, float elapsedTime )
{
	// Reset counters
	mTotalTriangles	= 0;
	mTotalVertices	= 0;

	// Record average fps
	static float fpsTime = 0.f;	
	fpsTime += elapsedTime;	
	if( fpsTime > 1.f )
	{
		static float fpsAccumulator = 0.f;
		static float frameCount = 0.f;
		
		float currentFps = DXUTGetFPS();
		fpsAccumulator += currentFps;						
		frameCount += 1.f;		
		
		mAvgFps = fpsAccumulator / frameCount;

		// Best and worst
		if( mBestFps < currentFps ) 
			mBestFps = currentFps;
		if( mWorstFps > currentFps && currentFps < 30.f)
			mWorstFps = currentFps; 		

		// Reset time
		fpsTime = 0.f;
	}	

	// Update the camera's position based on user input 
	mCamera.FrameMove( elapsedTime );	
}

LRESULT App::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
		case WM_CLOSE:			
			mAppExiting = true;
			LOG_INFO( "Shutting down application..." );
			break;		
	}

	bool pbNoFurtherProcessing;

	// Pass messages to dialog resource manager calls so GUI state is updated correctly
	pbNoFurtherProcessing = mDialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
	if( pbNoFurtherProcessing )
		return 0;

	// Pass messages to settings dialog if its active
	if( mSettingsDlg.IsActive() )
	{
		mSettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
		return 0;
	}

	pbNoFurtherProcessing = mAppDlg.MsgProc( hWnd, uMsg, wParam, lParam );
	if( pbNoFurtherProcessing )
		return 0;
	
	// Moved to client for this demo
	// mCamera.HandleMessages( hWnd, uMsg, wParam, lParam );

#ifdef INIT_CLIENT    
	MsgProcClient( hWnd, uMsg, wParam, lParam );
#endif

	return 0;
}

void App::OneTimeInit()
{
	// Store app window and instance
	mWndHandle		= DXUTGetHWND();
	mInstHandle		= DXUTGetHINSTANCE();		

	// Init GUI
	mSettingsDlg.Init( &mDialogResourceManager );	
	mAppDlg.Init( &mDialogResourceManager );
	mAppDlg.SetCallback( OnGUIEvent ); 	

	mCurrentGuiHeight = 0;
	mAppDlg.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 25, mCurrentGuiHeight, 125, 22 );
	mAppDlg.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 25, mCurrentGuiHeight += 24, 125, 22, VK_F3 );
	mAppDlg.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 25, mCurrentGuiHeight += 24, 125, 22, VK_F2 );
	mAppDlg.AddCheckBox( IDC_TOGGLEWIREFRAME, L"Wireframe", 25, mCurrentGuiHeight += 24, 125, 22, false );	
	mAppDlg.AddCheckBox( IDC_SHOWGUI, L"Hide gui (F1 to show)", 25, mCurrentGuiHeight += 24, 200, 22, false );	
	mAppDlg.AddStatic( IDC_CAMSPEEDTEXT, L"Cam speed:", 25, mCurrentGuiHeight +=24, 125, 22    );
	mAppDlg.AddSlider( IDC_SLIDERCAMSPEED, 25, mCurrentGuiHeight += 24, 125, 22, 1, 200, (int)mCamSpeed );

	CreateGridPoints( 100, 100, 10 );

	// Setup the camera's view parameters
	D3DXVECTOR3 vecEye( 0.f, 30.f, -200.f );
	D3DXVECTOR3 vecAt ( 0.f, 30.f, -1.f );
	mCamera.SetEnablePositionMovement( true );		
	mCamera.SetScalers( 0.01f, mCamSpeed );
	mCamera.SetRotateButtons( true, true, false, false );
	mCamera.SetViewParams( &vecEye, &vecAt );	
	mCamera.SetRotateButtons( true, false, false );		
}

//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void App::RenderText()
{
	D3DXMATRIX matWorld;
	D3DXMatrixScaling( &matWorld, 1.5f, 1.5f, 0.f );
	mSprite->SetTransform( &matWorld );

	CDXUTTextHelper txtHelper( mFont, mSprite, 15 );
	
	txtHelper.Begin();
	txtHelper.SetInsertionPos( 5, 5 );
	txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
	txtHelper.DrawTextLine( DXUTGetFrameStats( true ) );
	txtHelper.DrawTextLine( DXUTGetDeviceStats() );

	WCHAR temp[128];
	sprintf_s( mDebugString, "Triangles: %d, Vertices: %d", mTotalTriangles, mTotalVertices);
	txtHelper.DrawTextLine( ConvertString( (WCHAR*)temp, (char*) mDebugString ) );
	
	/*
	sprintf_s( debugString, "Camera pos x: %.2f, y: %.2f, z: %.2f Lookat x: %.2f, y: %.2f, z: %.2f yaw: %.2f pitch: %.2f", mCamera.GetEyePt()->x, mCamera.GetEyePt()->y, mCamera.GetEyePt()->z,
		mCamera.GetLookAtPt()->x, mCamera.GetLookAtPt()->y, mCamera.GetLookAtPt()->z, mCamera.GetYawAngle(), mCamera.GetPitchAngle() );
	mDebugStrings.push_back(debugString);
	*/

	for( size_t i = 0; i < mDebugStrings.size(); i++ )
	{
		WCHAR temp[128];
		txtHelper.DrawTextLine( ConvertString( (WCHAR*)temp, (char*) mDebugStrings.at(i).c_str() ) );
	}

	txtHelper.End();

	mDebugStrings.clear();
}

//--------------------------------------------------------------------------------------
// Returns true if a particular depth-stencil format can be created and used with
// an adapter format and backbuffer format combination.
BOOL IsDepthFormatOk( D3DFORMAT DepthFormat, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat )
{
	// Verify that the depth format exists
	HRESULT hr = DXUTGetD3D9Object()->CheckDeviceFormat( D3DADAPTER_DEFAULT, 
		D3DDEVTYPE_HAL, AdapterFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, DepthFormat );
	if( FAILED( hr ) ) return FALSE;

	// Verify that the backbuffer format is valid
	hr = DXUTGetD3D9Object()->CheckDeviceFormat( D3DADAPTER_DEFAULT, 
		D3DDEVTYPE_HAL, AdapterFormat, D3DUSAGE_RENDERTARGET, D3DRTYPE_SURFACE, BackBufferFormat );
	if( FAILED( hr ) ) return FALSE;

	// Verify that the depth format is compatible
	hr = DXUTGetD3D9Object()->CheckDepthStencilMatch( D3DADAPTER_DEFAULT, 
		D3DDEVTYPE_HAL, AdapterFormat, BackBufferFormat, DepthFormat );

	return SUCCEEDED( hr );
}

//--------------------------------------------------------------------------------------
// Rejects any D3D9 devices that aren't acceptable to the app by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D9DeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat,
									 D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
	// Skip backbuffer formats that don't support alpha blending
	IDirect3D9* pD3D = DXUTGetD3D9Object();
	if( FAILED( pD3D->CheckDeviceFormat( pCaps->AdapterOrdinal, pCaps->DeviceType,
		AdapterFormat, D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING, D3DRTYPE_TEXTURE, BackBufferFormat ) ) )
		return false;

	// No fallback defined by this app, so reject any device that 
	// doesn't support at least ps2.0
	if( pCaps->PixelShaderVersion < D3DPS_VERSION( 2, 0 ) )
		return false;

	// Must support stencil buffer
	if( !IsDepthFormatOk( D3DFMT_D24S8, AdapterFormat, BackBufferFormat ) && 
		!IsDepthFormatOk( D3DFMT_D24X4S4, AdapterFormat, BackBufferFormat ) &&
		!IsDepthFormatOk( D3DFMT_D15S1,	AdapterFormat, BackBufferFormat ) && 
		!IsDepthFormatOk( D3DFMT_D24FS8, AdapterFormat, BackBufferFormat ) )
		return false;

	return true;
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
	if( pDeviceSettings->ver == DXUT_D3D9_DEVICE )
	{
		IDirect3D9* pD3D = DXUTGetD3D9Object();
		D3DCAPS9 Caps;
		pD3D->GetDeviceCaps( pDeviceSettings->d3d9.AdapterOrdinal, pDeviceSettings->d3d9.DeviceType, &Caps );

		// If device doesn't support HW T&L or doesn't support 1.1 vertex shaders in HW 
		// then switch to SWVP.
		if( ( Caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) == 0 ||
			Caps.VertexShaderVersion < D3DVS_VERSION( 1, 1 ) )
		{
			pDeviceSettings->d3d9.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
		}
#ifndef ENABLE_VSYNC
		pDeviceSettings->d3d9.pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		gApp->GetD3DSettingsDialog()->GetDialogControl()->GetComboBox( DXUTSETTINGSDLG_PRESENT_INTERVAL )->SetEnabled( false );
#endif

		// Debugging vertex shaders requires either REF or software vertex processing 
		// and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
		if( pDeviceSettings->d3d9.DeviceType != D3DDEVTYPE_REF )
		{
			pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
			pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
			pDeviceSettings->d3d9.BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
		}
#endif
#ifdef DEBUG_PS
		pDeviceSettings->d3d9.DeviceType = D3DDEVTYPE_REF;
#endif
	}

	// This sample requires a stencil buffer.
	if( IsDepthFormatOk( D3DFMT_D24S8,
		pDeviceSettings->d3d9.AdapterFormat,
		pDeviceSettings->d3d9.pp.BackBufferFormat ) )
		pDeviceSettings->d3d9.pp.AutoDepthStencilFormat = D3DFMT_D24S8;
	else if( IsDepthFormatOk( D3DFMT_D24X4S4,
		pDeviceSettings->d3d9.AdapterFormat,
		pDeviceSettings->d3d9.pp.BackBufferFormat ) )
		pDeviceSettings->d3d9.pp.AutoDepthStencilFormat = D3DFMT_D24X4S4;
	else if( IsDepthFormatOk( D3DFMT_D24FS8,
		pDeviceSettings->d3d9.AdapterFormat,
		pDeviceSettings->d3d9.pp.BackBufferFormat ) )
		pDeviceSettings->d3d9.pp.AutoDepthStencilFormat = D3DFMT_D24FS8;
	else if( IsDepthFormatOk( D3DFMT_D15S1,
		pDeviceSettings->d3d9.AdapterFormat,
		pDeviceSettings->d3d9.pp.BackBufferFormat ) )
		pDeviceSettings->d3d9.pp.AutoDepthStencilFormat = D3DFMT_D15S1;
	
	pDeviceSettings->d3d9.pp.EnableAutoDepthStencil = TRUE;
	// For the first device created if its a REF device, optionally display a warning dialog box
	static bool s_bFirstTime = true;
	if( s_bFirstTime )
	{
		s_bFirstTime = false;
		if( ( DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF ) ||
			( DXUT_D3D10_DEVICE == pDeviceSettings->ver &&
			pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE ) )
			DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
	}

	return true;
}


HRESULT App::OnCreateDevice(IDirect3DDevice9* device, const D3DSURFACE_DESC* backBufferSurfaceDesc)
{
	LOG_DEBUG( "OnCreateDevice" );
	mDevice = device;

	HRESULT hr;

	V_RETURN( mDialogResourceManager.OnD3D9CreateDevice( mDevice ) );
	V_RETURN( mSettingsDlg.OnD3D9CreateDevice( mDevice ) );

	V_RETURN( D3DXCreateFont( mDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
		L"Arial", &mFont ) );	

	V_RETURN( D3DXCreateSprite( mDevice, &mSprite ) );	
	V_RETURN( D3DXCreateLine( mDevice, &mLine ) );	

	CreateGlobalVariables(mDevice);
	CreateAllVertexDeclarations( mDevice );

	return hr;	
}

HRESULT App::OnResetDevice(IDirect3DDevice9* device, const D3DSURFACE_DESC* backBufferSurfaceDesc)
{
	mDevice = device;

	LOG_DEBUG( "App::OnResetDevice" );
	float fAspectRatio = backBufferSurfaceDesc->Width / ( FLOAT )backBufferSurfaceDesc->Height;

	HRESULT hr = S_OK;

	V_RETURN( mDialogResourceManager.OnD3D9ResetDevice() );
	V_RETURN( mSettingsDlg.OnD3D9ResetDevice() );

	if( mFont ) V_RETURN( mFont->OnResetDevice() );
	if( mSprite ) V_RETURN ( mSprite->OnResetDevice() ) ;
	if( mLine ) V_RETURN ( mLine->OnResetDevice() ) ;	

	ResetGlobalVariables(device);

	mAppDlg.SetLocation( backBufferSurfaceDesc->Width - 170, 20 );
	mAppDlg.SetSize( 170, 300 );		

	// Set projections, could be lost when a device has been set
	mScreenWidth = backBufferSurfaceDesc->Width;
	mScreenHeight = backBufferSurfaceDesc->Height;
	mCamera.SetProjParams( FOV, fAspectRatio, ZNEARPLANE, ZFARPLANE );	

	// Enable specular material and put in some ambient light
	mDevice->SetRenderState( D3DRS_SPECULARMATERIALSOURCE, D3DMCS_MATERIAL );
	mDevice->SetRenderState( D3DRS_SPECULARENABLE, TRUE );
	mDevice->SetRenderState( D3DRS_NORMALIZENORMALS, true );
	mDevice->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_XRGB(50, 50, 50) );	
	
	// Put a directional light by default
	mDevice->SetLight( 0, &mDirLight );
	mDevice->LightEnable( 0, true );	

#ifdef CONFINE_MOUSE
	ConfineMouse();
#endif



	return hr;
}

void App::MouseProc( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos )
{
	mRightButtonDown = bRightButtonDown;
	
	// Send to client
	MouseProcClient( bLeftButtonDown, bRightButtonDown, bMiddleButtonDown, bSideButton1Down, bSideButton2Down, nMouseWheelDelta, xPos, yPos );
}

void App::KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown )
{
	if( bKeyDown )
	{
		switch( nChar )
		{	
		case VK_F1:
			mAppDlg.GetCheckBox( IDC_SHOWGUI )->SetChecked( !mAppDlg.GetCheckBox( IDC_SHOWGUI )->GetChecked() );			
			mShowGui = !mShowGui;
			break;
		case VK_ESCAPE:			
			mAppExiting = true;
			break;
		}
	}

	// Send to client
	KeyboardProcClient( nChar, bKeyDown, bAltDown );
}

void App::GUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl )
{
	switch( nControlID )
	{
	case IDC_TOGGLEFULLSCREEN:
		DXUTToggleFullScreen(); 
		break;
	case IDC_TOGGLEREF:
		DXUTToggleREF(); 
		break;
	case IDC_CHANGEDEVICE:
		mSettingsDlg.SetActive( !mSettingsDlg.IsActive() ); 
		break;
	case IDC_TOGGLEWIREFRAME:
		if( (( CDXUTCheckBox* )pControl)->GetChecked() )
		{
			mDevice->SetRenderState( D3DRS_CULLMODE,			D3DCULL_NONE );
			mDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,	false );
			mDevice->SetRenderState( D3DRS_FILLMODE,			D3DFILL_WIREFRAME );
		}
		else
		{
			mDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			mDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
		}
		break;
	case IDC_SLIDERCAMSPEED:	
		mCamera.SetScalers( 0.01f, (float)mAppDlg.GetSlider( IDC_SLIDERCAMSPEED )->GetValue() );
		break;
	}
	
	GUIEventClient( nEvent, nControlID, pControl );
}

void App::OnLostDevice()
{
	LOG_DEBUG( "OnLostDevice()" );

	mDialogResourceManager.OnD3D9LostDevice();
	mSettingsDlg.OnD3D9LostDevice();
	if( mFont ) mFont->OnLostDevice();	
	if( mSprite ) mSprite->OnLostDevice();	
	if( mLine ) mLine->OnLostDevice();	
	LostGlobalVariables();
}

void App::OnDestroyDevice()
{
	LOG_DEBUG( "OnDestroyDevice()" );

	mDialogResourceManager.OnD3D9DestroyDevice();
	mSettingsDlg.OnD3D9DestroyDevice();
	SAFE_RELEASE( mFont );			
	SAFE_RELEASE( mSprite );	
	SAFE_RELEASE( mLine );
	DestroyGlobalVariables();
	DestroyAllVertexDeclarations();	
}

D3DXVECTOR3 App::GetRayTo(int x, int y)
{
	D3DXVECTOR3 rayForward;

	D3DXVECTOR3	rayFrom = *mCamera.GetEyePt();
	D3DXVec3Subtract( &rayForward, mCamera.GetLookAtPt(), mCamera.GetEyePt() );			
	D3DXVec3Normalize( &rayForward, &rayForward );

	float farPlane = ZFARPLANE;
	rayForward *= farPlane;	
	
	D3DXVECTOR3 vertical(0.f, 1.f, 0.f);
	D3DXVECTOR3 horizontal;
	
	D3DXVec3Cross( &horizontal, &rayForward, &D3DXVECTOR3(-vertical.x, -vertical.y, -vertical.z) );
	D3DXVec3Normalize( &horizontal, &horizontal );	
	D3DXVec3Cross( &vertical, &horizontal, &D3DXVECTOR3(-rayForward.x, -rayForward.y, -rayForward.z) );	
	D3DXVec3Normalize( &vertical, &vertical );

	float tanfov = tanf(0.5f * FOV);
	float aspect = mScreenHeight / (float) mScreenWidth;

	horizontal *= 2.f * farPlane * tanfov;
	vertical *= 2.f * farPlane * tanfov;

	if (aspect<1)
	{
		horizontal /= aspect;
	} 
	else
	{
		vertical *= aspect;
	}

	D3DXVECTOR3 rayToCenter = rayFrom + rayForward;
	D3DXVECTOR3 dHor = horizontal * 1.f / float(mScreenWidth);
	D3DXVECTOR3 dVert = vertical * 1.f / float(mScreenHeight);		
	
	D3DXVECTOR3 rayTo = rayToCenter - 0.5f * horizontal + 0.5f * vertical;	
	rayTo += (float)x * dHor;
	rayTo -= (float)y * dVert;		

	return rayTo;
}

void App::RenderAxis()
{
	// Make sure render axis shows up all the time
	// by setting third column to all 0.0 to make final Z = 0
	D3DXMATRIX oldProj, newProj;
	mDevice->GetTransform( D3DTS_PROJECTION, &oldProj );
	newProj = oldProj;
	newProj._13 = 0.f;
	newProj._23 = 0.f;
	newProj._33 = 0.f;
	newProj._43 = 0.f;
	//mDevice->SetTransform( D3DTS_PROJECTION, &newProj );	

	float axisLength = 10.0f;
	float letterSize = 2.0f;

	VertexPos vertices[22];

	// X axis
	vertices[0].position = D3DXVECTOR3(-axisLength, 0.0f, 0.0f);
	vertices[1].position = D3DXVECTOR3(axisLength, 0.0f, 0.0f);

	// X letter
	vertices[2].position = D3DXVECTOR3(axisLength, letterSize, 0.0f);				// Upper right
	vertices[3].position = D3DXVECTOR3(axisLength - letterSize, 0.2f, 0.0f);
	vertices[4].position = D3DXVECTOR3(axisLength - letterSize, letterSize, 0.0f);	// Upper left
	vertices[5].position = D3DXVECTOR3(axisLength, 0.2f, 0.0f);	

	// Y axis
	vertices[6].position = D3DXVECTOR3(0.0f, -axisLength, 0.0f);
	vertices[7].position = D3DXVECTOR3(0.0f, axisLength, 0.0f);

	// Y letter
	vertices[8].position = D3DXVECTOR3(letterSize, axisLength, 0.0f);				// Verticle
	vertices[9].position = D3DXVECTOR3(letterSize, axisLength - letterSize, 0.0f);
	vertices[10].position = D3DXVECTOR3(letterSize, axisLength, 0.0f);				// Right diagonal
	vertices[11].position = D3DXVECTOR3(letterSize + letterSize, axisLength + letterSize, 0.0f);
	vertices[12].position = D3DXVECTOR3(letterSize, axisLength, 0.0f);				// Left diagonal
	vertices[13].position = D3DXVECTOR3(0.2f, axisLength + letterSize, 0.0f);

	// Z axis
	vertices[14].position = D3DXVECTOR3(0.0f, 0.0f, -axisLength);
	vertices[15].position = D3DXVECTOR3(0.0f, 0.0f, axisLength);

	// Z letter
	vertices[16].position = D3DXVECTOR3(0.0f, 0.2f, axisLength);					// Bottom horizontal
	vertices[17].position = D3DXVECTOR3(0.0f, 0.2f, axisLength - letterSize);
	vertices[18].position = D3DXVECTOR3(0.0f, 0.2f, axisLength);					// Diagonal 
	vertices[19].position = D3DXVECTOR3(0.0f, 0.2f + letterSize, axisLength - letterSize);
	vertices[20].position = D3DXVECTOR3(0.0f, 0.2f + letterSize, axisLength - letterSize); // Top horizontal 
	vertices[21].position = D3DXVECTOR3(0.0f, 0.2f + letterSize, axisLength);

	D3DXMATRIX matCoordinate;
	D3DXMatrixIdentity(&matCoordinate);
	
	D3DMATERIAL9 oldMaterial;	
	DWORD oldAmbient;
	mDevice->GetMaterial( &oldMaterial );
	mDevice->GetRenderState( D3DRS_AMBIENT, &oldAmbient );
	
	// Disable depth buffering so Axis always shows up		
	mDevice->SetRenderState( D3DRS_ZENABLE, false );	
	mDevice->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_RGBA(255, 255, 255, 255) );	
	
	mDevice->SetTransform( D3DTS_WORLD, &matCoordinate );	
	mDevice->SetFVF(D3DFVF_VERTEXPOSITION);
 
	Material coordinateMat;		
	coordinateMat.SetAmbientColor( Color::Red );
	mDevice->SetMaterial( &coordinateMat );		
	mDevice->DrawPrimitiveUP( D3DPT_LINELIST, 3, (LPVOID) &vertices[0], sizeof(VertexPos)  );
	
	coordinateMat.SetAmbientColor( Color::Green );
	mDevice->SetMaterial( &coordinateMat );		
	mDevice->DrawPrimitiveUP( D3DPT_LINELIST, 4, (LPVOID) &vertices[6], sizeof(VertexPos)  );

	coordinateMat.SetAmbientColor( Color::Blue ) ;	
	mDevice->SetMaterial( &coordinateMat );		
	mDevice->DrawPrimitiveUP( D3DPT_LINELIST, 4, (LPVOID) &vertices[14], sizeof(VertexPos)  );

	// Restore states
	mDevice->SetRenderState( D3DRS_ZENABLE, true );	
	mDevice->SetRenderState( D3DRS_AMBIENT, oldAmbient );			
	mDevice->SetMaterial( &oldMaterial );		

	mDevice->SetTransform( D3DTS_PROJECTION, &oldProj );
}

void App::RenderGrid()
{
	D3DMATERIAL9 oldMaterial;	

	mDevice->GetMaterial( &oldMaterial );

	Material whiteMat;		
	whiteMat.SetDiffuseColor( Color::White );
	whiteMat.SetAmbientColor( Color::White );
	mDevice->SetMaterial( &whiteMat );		

	D3DXMATRIX matWorld;
	D3DXMatrixIdentity(&matWorld);
	mDevice->SetTransform( D3DTS_WORLD, &matWorld );
	mDevice->SetFVF(D3DFVF_VERTEXPOSITION);
	mDevice->DrawPrimitiveUP(D3DPT_LINELIST, mNumGridPrimitives, (LPVOID) mGridPoints, sizeof(VertexPos));	

	mDevice->SetMaterial( &oldMaterial );		
}

void App::CreateGridPoints(int numRows, int numCols, float unitSize)
{
	// Add 4 more points to accommodate the last 2 lines to close the grid
	int numGridPoints = ((numRows + numCols) * 2) + 4;

	// Calculate grid maximum width and height
	float width = (float)numCols * unitSize;	
	float depth = (float)numRows * unitSize;
	
	// Use to center the grid
	D3DXVECTOR3 center(-width / 2.f,0, -depth / 2.f);	

	// Store the number of primitive. Add 2 to accomodate the additional 2 lines
	mNumGridPrimitives = numRows + numCols + 2;

	mGridPoints = new D3DXVECTOR3[ numGridPoints ];	
	int index = 0;

	// Create the horizontal lines
	for(float i = 0; i < numRows; ++i)
	{
		D3DXVECTOR3 s, e;		
		
		s.x = 0.f;
		s.z = i * unitSize;
		s.y = 0.f;
		mGridPoints[index++] = s;

		e.x = width;
		e.z = i * unitSize;
		e.y = 0.f;
		mGridPoints[index++] = e;
	}
	mGridPoints[index++] = D3DXVECTOR3(0.f, 0.f, numRows * unitSize);
	mGridPoints[index++] = D3DXVECTOR3(width, 0.f, numRows * unitSize);

	// Create the vertical lines
	for(float i = 0; i < numCols; ++i)
	{
		D3DXVECTOR3 s, e;		
		
		s.x = i * unitSize;
		s.z = depth;
		s.y = 0.f;
		mGridPoints[index++] = s;

		e.x = i * unitSize;
		e.z = 0.f;
		e.y = 0.f;
		mGridPoints[index++] = e;
	}
	mGridPoints[index++] = D3DXVECTOR3(numCols * unitSize, 0.f, 0.f);
	mGridPoints[index++] = D3DXVECTOR3(numCols * unitSize, 0.f, depth);

	// Center the grid
	for(int i = 0; i < ((numRows + numCols) * 2) + 4; i++ )
	{
		D3DXMATRIX T;
		D3DXMatrixTranslation(&T, center.x, center.y, center.z);
		D3DXVec3TransformCoord(&mGridPoints[i], &mGridPoints[i], &T);
	}
}

void App::ConfineMouse()
{
	RECT rcClip;
	RECT rcOldClip; 
	GetClipCursor( &rcOldClip ); 
	GetWindowRect( DXUTGetHWND(), &rcClip ); 
	ClipCursor(&rcClip); 
	// Center cursor
	int x = rcClip.left + rcClip.right / 2;
	int y = rcClip.top + rcClip.bottom / 2;
	SetCursorPos( x, y );		
}

//--------------------------------------------------------------------------------------
// Create any D3D9 resources that will live through a device reset (D3DPOOL_MANAGED)
// and aren't tied to the back buffer size
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D9CreateDevice( IDirect3DDevice9* device, const D3DSURFACE_DESC* backBufferSurfaceDesc,
									void* pUserContext )
{
	LOG_DEBUG( "Begin OnD3D9CreateDevice()" );
	gApp->OnCreateDevice( device, backBufferSurfaceDesc );

#ifdef INIT_CLIENT
	LOG_DEBUG( "Calling CreateClient()..." );
	gApp->CreateClient();
#endif	
	
	LOG_DEBUG( "End OnD3D9CreateDevice()" );

	return S_OK;
}

//--------------------------------------------------------------------------------------
// Create any D3D9 resources that won't live through a device reset (D3DPOOL_DEFAULT) 
// or that are tied to the back buffer size 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D9ResetDevice( IDirect3DDevice9* device,
								   const D3DSURFACE_DESC* backBufferSurfaceDesc, void* pUserContext )
{
	LOG_DEBUG( "Begin OnD3D9ResetDevice()" );

	gApp->OnResetDevice(device, backBufferSurfaceDesc);	
	
#ifdef INIT_CLIENT
	LOG_DEBUG( "Calling ResetClient()..." );
	gApp->ResetClient();
#endif	

	LOG_DEBUG( "End OnD3D9ResetDevice()" );

	return S_OK;	
}

//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double time, float elapsedTime, void* pUserContext )
{
	gApp->OnFrameMove( time, elapsedTime );	

#ifdef INIT_CLIENT
	gApp->UpdateClient( time, elapsedTime );	
#endif		
}

//--------------------------------------------------------------------------------------
// Render the scene using the D3D9 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D9FrameRender( IDirect3DDevice9* device, double time, float elapsedTime, void* pUserContext )
{
	HRESULT hr;
	D3DXMATRIXA16 mWorld;
	D3DXMATRIXA16 mView;
	D3DXMATRIXA16 mProj;
	D3DXMATRIXA16 mWorldViewProjection;

	// If the settings dialog is being shown, then render it instead of rendering the app's scene
	if( gApp->GetD3DSettingsDialog()->IsActive() )
	{
		gApp->GetD3DSettingsDialog()->OnRender( elapsedTime );
		return;
	}		

	// Clear the render target and the zbuffer 
	V( device->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_ARGB( 0, 0, 0, 0 ), 1.0f, 0 ) );

	// Render the scene
	if( SUCCEEDED( device->BeginScene() ) )
	{
		gApp->OnFrameRender( time, elapsedTime );	

#ifdef INIT_CLIENT		
		// Call render client
		gApp->RenderClient( time, elapsedTime );
#endif

#ifdef DRAW_AXIS
		gApp->RenderAxis();
#endif

		if( !gApp->GetAppDlg()->GetCheckBox( IDC_SHOWGUI )->GetChecked() )
		{
			gApp->GetAppDlg()->OnRender( elapsedTime );
	
			DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" ); // These events are to help PIX identify what the code is doing
			gApp->RenderText();
			DXUT_EndPerfEvent();
		}

		V( device->EndScene() );
	}
}

//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
						 void* pUserContext )
{
	return gApp->MsgProc( hWnd, uMsg, wParam, lParam );		
}

void CALLBACK OnMouseProc( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos, void* pUserContext )
{
	gApp->MouseProc( bLeftButtonDown, bRightButtonDown, bMiddleButtonDown, bSideButton1Down, bSideButton2Down, nMouseWheelDelta, xPos, yPos );
}

//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	gApp->KeyboardProc( nChar, bKeyDown, bAltDown );
}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
	gApp->GUIEvent( nEvent, nControlID, pControl );
}

//--------------------------------------------------------------------------------------
// Release D3D9 resources created in the OnD3D9ResetDevice callback 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D9LostDevice( void* pUserContext )
{
	if( gApp )
	{
		gApp->OnLostDevice();
		LOG_DEBUG( "Calling LostClient()..." );
#ifdef INIT_CLIENT
		gApp->LostClient();		
#endif
	}
}

//--------------------------------------------------------------------------------------
// Release D3D9 resources created in the OnD3D9CreateDevice callback 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D9DestroyDevice( void* pUserContext )
{
	if( gApp )
	{
		gApp->OnDestroyDevice();
#ifdef INIT_CLIENT
		gApp->DestroyClient();
#endif
	}
	
	if( gApp->IsAppExiting() )
	{
		SAFE_DELETE(gApp);
		LOG_DEBUG( "Application terminated." );		
	}
}