#include "dxut.h"
#include "Camera.h"

Camera::Camera(void)
{
	LOG_DEBUG( "Camera constructor called." );	

	m_vMinBoundary = D3DXVECTOR3( -175.0f, 2.5f, -175.f );
	m_vMaxBoundary = D3DXVECTOR3( 175, 150.f, 175.f );
	m_bClipToBoundary = false;
	m_fCameraPitchAngle = m_fCameraYawAngle = 0.5f;	
}

Camera::~Camera(void)
{
	LOG_DEBUG( "Camera destructor called." );
}

void Camera::FrameMove( float elapsedTime )
{
	CFirstPersonCamera::FrameMove( elapsedTime );
	
	CalculateFrustumPlanes();
}

void Camera::CalculateFrustumPlanes()
{
	// Get combined matrix
	D3DXMATRIX matComb;
	D3DXMatrixMultiply(&matComb, &m_mView, &m_mProj);

	// Left clipping plane
	mPlanes[0].a = matComb._14 + matComb._11; 
	mPlanes[0].b = matComb._24 + matComb._21; 
	mPlanes[0].c = matComb._34 + matComb._31; 
	mPlanes[0].d = matComb._44 + matComb._41;

	// Right clipping plane 
	mPlanes[1].a = matComb._14 - matComb._11; 
	mPlanes[1].b = matComb._24 - matComb._21; 
	mPlanes[1].c = matComb._34 - matComb._31; 
	mPlanes[1].d = matComb._44 - matComb._41;

	// Top clipping plane 
	mPlanes[2].a = matComb._14 - matComb._12; 
	mPlanes[2].b = matComb._24 - matComb._22; 
	mPlanes[2].c = matComb._34 - matComb._32; 
	mPlanes[2].d = matComb._44 - matComb._42;

	// Bottom clipping plane 
	mPlanes[3].a = matComb._14 + matComb._12; 
	mPlanes[3].b = matComb._24 + matComb._22; 
	mPlanes[3].c = matComb._34 + matComb._32; 
	mPlanes[3].d = matComb._44 + matComb._42;	  

	// Near clipping plane 
	mPlanes[4].a = matComb._14 + matComb._13; 
	mPlanes[4].b = matComb._24 + matComb._23; 
	mPlanes[4].c = matComb._34 + matComb._33; 
	mPlanes[4].d = matComb._44 + matComb._43;

	// Far clipping plane 
	mPlanes[5].a = matComb._14 - matComb._13; 
	mPlanes[5].b = matComb._24 - matComb._23; 
	mPlanes[5].c = matComb._34 - matComb._33; 
	mPlanes[5].d = matComb._44 - matComb._43; 

	//Normalize planes
	for(int i=0;i<6;i++)
		D3DXPlaneNormalize(&mPlanes[i], &mPlanes[i]);
}

bool Camera::FrustumSphereCheck(D3DXVECTOR3 &center, float radius)
{
	// Make sure radius is in frustum
	for( int i = 0; i < 6; i++ ) 
	{
		if(D3DXPlaneDotCoord(&mPlanes[i], &center) < -radius)
			return false;
	}
	return true;
}

bool Camera::FrustumBoxCheck( D3DXVECTOR3 &center, float size )
{
	// Make sure at least one point is completely in frustum
	for( int i = 0; i < 6; i++ )
	{
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x-size, center.y-size, center.z-size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x+size, center.y-size, center.z-size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x-size, center.y+size, center.z-size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x+size, center.y+size, center.z-size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x-size, center.y-size, center.z+size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x+size, center.y-size, center.z+size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x-size, center.y+size, center.z+size)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(center.x+size, center.y+size, center.z+size)) >= 0.0f)
			continue;

		return false;
	}
	return true;
}

bool Camera::FrustumBoxCheck( D3DXVECTOR3 &min, D3DXVECTOR3 &max )
{
	// Make sure at least one point is completely in frustum
	for( int i = 0; i < 6; i++ )
	{
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( min.x, min.y, min.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( max.x, min.y, min.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( min.x, max.y, min.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( max.x, max.y, min.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( min.x, min.y, max.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( max.x, min.y, max.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( min.x, max.y, max.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( max.x, max.y, max.z ) ) >= 0.0f )
			continue;

		return false;
	}
	return true;
}

void Camera::CalculateFrustumCorners()
{
	// Calculate the frustum corners in world space

	D3DXVECTOR3 vZ = m_vLookAt - m_vEye;
	D3DXVec3Normalize( &vZ, &vZ );

	D3DXVECTOR3 vX;
	D3DXVec3Cross( &vX, GetWorldUp(), &vZ );
	D3DXVec3Normalize( &vX, &vX );

	D3DXVECTOR3 vY;
	D3DXVec3Cross( &vY, &vZ, &vX );
	
	float fNearPlaneHeight = tanf(m_fFOV * 0.5f) * m_fNearPlane;
	float fNearPlaneWidth = fNearPlaneHeight * m_fAspect;

	float fFarPlaneHeight = tanf(m_fFOV * 0.5f) * m_fFarPlane;
	float fFarPlaneWidth = fFarPlaneHeight * m_fAspect;

	D3DXVECTOR3 vNearPlaneCenter = m_vEye + vZ * m_fNearPlane;
	D3DXVECTOR3 vFarPlaneCenter = m_vEye + vZ * m_fFarPlane;

	// Near
	mFrustumPoints[0]=D3DXVECTOR3(vNearPlaneCenter - vX * fNearPlaneWidth - vY * fNearPlaneHeight); // LR
	mFrustumPoints[1]=D3DXVECTOR3(vNearPlaneCenter - vX * fNearPlaneWidth + vY * fNearPlaneHeight); // UR
	mFrustumPoints[2]=D3DXVECTOR3(vNearPlaneCenter + vX * fNearPlaneWidth + vY * fNearPlaneHeight); // UL
	mFrustumPoints[3]=D3DXVECTOR3(vNearPlaneCenter + vX * fNearPlaneWidth - vY * fNearPlaneHeight); // LL

	// Far
	mFrustumPoints[4]=D3DXVECTOR3(vFarPlaneCenter - vX * fFarPlaneWidth - vY * fFarPlaneHeight); // LR
	mFrustumPoints[5]=D3DXVECTOR3(vFarPlaneCenter - vX * fFarPlaneWidth + vY * fFarPlaneHeight); // UR
	mFrustumPoints[6]=D3DXVECTOR3(vFarPlaneCenter + vX * fFarPlaneWidth + vY * fFarPlaneHeight); // UL
	mFrustumPoints[7]=D3DXVECTOR3(vFarPlaneCenter + vX * fFarPlaneWidth - vY * fFarPlaneHeight); // LL

	// calculate center of points
	D3DXVECTOR3 vCenter(0,0,0);
	for( int i = 0 ; i < 8; i++ ) 
	{
		vCenter += mFrustumPoints[i];
	}
	vCenter /= 8;

	// To avoid edge problems, scale the frustum so
    // that it's at least a few pixels larger	
	float fScale = 1.1f;
	// for each point
	for(int i = 0; i < 8; i++ )
	{
		// scale by adding offset from center
		mFrustumPoints[i] += (mFrustumPoints[i]-vCenter)*(fScale-1);
	}
}

	
ChaseCamera::ChaseCamera()
{
	mMass = 50.0f;
	mDamping = 600.0f;
	mStiffness = 800.0f;
	mLookAtOffset = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	mTargetPositionOffset = D3DXVECTOR3( 0.0f, 10.0f, 15.0f );
	Reset();
}

void ChaseCamera::UpdateWorldPositions()
{
	D3DXMATRIX mat;
	D3DXMatrixIdentity( &mat );
	D3DXVECTOR3 vUp( 0.0f, 1.0f, 0.0f );
	D3DXVECTOR3 vRight;
	D3DXVec3Cross( &vRight, &vUp, &mChaseDirection );	

	// right vector
	mat._11 = vRight.x;
	mat._12 = vRight.y;
	mat._13 = vRight.z;

	// up vector
	mat._21 = vUp.x;
	mat._22 = vUp.y;
	mat._23 = vUp.z;

	// forward vector
	mat._31 = mChaseDirection.x;
	mat._32 = mChaseDirection.y;
	mat._33 = mChaseDirection.z;


	D3DXVECTOR3 vTargetPosTransform;
	D3DXVec3TransformNormal( &vTargetPosTransform, &mTargetPositionOffset, &mat );

	D3DXVECTOR3 vLookAtTransform;
	D3DXVec3TransformNormal( &vLookAtTransform, &mLookAtOffset, &mat );

	mTargetPosition = mChasePosition + vTargetPosTransform;
	m_vLookAt = mChasePosition + vLookAtTransform;	
}

void ChaseCamera::FrameMove( FLOAT elapsedTime )
{
	UpdateWorldPositions();

	D3DXVECTOR3 vStretch = m_vEye - mTargetPosition;
	D3DXVECTOR3 vForce = ( -mStiffness * vStretch ) - ( mDamping * m_vVelocity ); 

	// Apply acceleration
	D3DXVECTOR3 vAcc = vForce / mMass;
	m_vVelocity += vAcc * elapsedTime;            

	// Apply velocity
	m_vEye += m_vVelocity * elapsedTime;

	// Update the view matrix
	D3DXMATRIX mCameraRot;
	D3DXMatrixRotationYawPitchRoll( &mCameraRot, m_fCameraYawAngle, m_fCameraPitchAngle, 0 );

	// Transform vectors based on camera's rotation matrix
	D3DXVECTOR3 vWorldUp;
	D3DXVECTOR3 vLocalUp = D3DXVECTOR3( 0, 1, 0 );    
	D3DXVec3TransformCoord( &vWorldUp, &vLocalUp, &mCameraRot );

	D3DXMatrixLookAtLH( &m_mView, &m_vEye, &m_vLookAt, &vLocalUp );
	D3DXMatrixInverse( &m_mCameraWorld, NULL, &m_mView );

	CalculateFrustumPlanes();
}

void ChaseCamera::Reset()
{
	UpdateWorldPositions();

	m_vVelocity = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	m_vEye = mTargetPosition;
}
